class InvalidDataframeType(ValueError):

    def __init__(self):
        self.message = 'Invalid dataframe input supplied, make sure your input is a valid pandas.DataFrame'

    def __str__(self):
        return self.message



import re
from bank_statement_reader.BaseBankStatementReport import BankStatementReport
from bank_statement_reader.exceptions.InprocessibleBankStatementSupplied import UnprocessableBankStatementSupplied
from datetime import datetime


class AccessBankStatement(BankStatementReport):
    version_one: bool = False
    version_two: bool = False

    def __init__(self, pdf_directory, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/access/access_version_one.pdf"  # version one
            # pdf_directory = "pdfs/access/access_version_two.pdf"  # version two
        self.pdf_directory = pdf_directory
        super().__init__(password='', pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary,
                         bank_name='access')

    def get_statement_period(self, data: str | dict = None):
        if isinstance(data, str) and self.version_one:
            return super().get_statement_period(data)
        if isinstance(data, dict) and self.version_two:
            date_string = data['summary_statement_for']
            # Split the date string into start and end date parts
            start_date_str, end_date_str = date_string.split(" to ")
            # Define the input and output date formats
            input_date_format = "%A, %B %d, %Y"
            output_date_format = "%Y-%m-%d"
            # Convert start and end date strings to datetime objects
            start_date = datetime.strptime(start_date_str, input_date_format)
            end_date = datetime.strptime(end_date_str, input_date_format)
            # Format the datetime objects as strings in the desired output format
            formatted_start_date = start_date.strftime(output_date_format)
            formatted_end_date = end_date.strftime(output_date_format)
            return {'from_date': formatted_start_date, 'to_date': formatted_end_date}
        return {'from_date': None, 'to_date': None}

    def get_account_number(self, data):
        if self.version_one:
            return super().get_account_number(data)
        if self.version_two and isinstance(data, dict):
            return data['account_number']
        return None

    def get_account_name(self, data: str | dict = None):
        if self.version_one:
            return super().get_account_name(data)
        if self.version_two and isinstance(data, dict):
            return data['account_name']
        return None

    def get_opening_balance(self, data):
        if self.version_one:
            return super().get_opening_balance(data)
        if self.version_two and isinstance(data, dict):
            return self.convert_to_money(data['opening_balance'])
        return None

    def get_closing_balance(self, data):
        if self.version_one:
            return super().get_closing_balance(data)
        if self.version_two and isinstance(data, dict):
            return self.convert_to_money(data['closing_balance'])
        return None

    def get_total_deposit(self, data):
        if self.version_one:
            return super().get_total_deposit(data)
        if self.version_two and isinstance(data, dict):
            return self.convert_to_money(data['total_lodgement'])
        return None

    def get_total_withdrawal(self, data):
        if self.version_one:
            return super().get_total_withdrawal(data)
        if self.version_two and isinstance(data, dict):
            return self.convert_to_money(data['total_withdrawals'])
        return None

    def get_transactions_table_header_mapping(self):
        if self.version_one:
            return {
                'date': 'Date',
                'transaction_details': 'Transaction Details',
                'reference': 'Reference',
                'value_date': 'Value Date',
                'withdrawals': 'Withdrawals',
                'lodgements': 'Lodgements',
                'balance': 'Balance'
            }
        if self.version_two:
            return {
                'posted_date': 'Posted Date',
                'value_date': 'Value Date',
                'description': 'Description',
                'debit': 'Debit',
                'credit': 'Credit',
                'balance': 'Balance'
            }

    def get_transactions_table_headers(self, reader):
        return super().get_transactions_table_headers(reader)

    def pad_header_with_unknown(self, rows, headers):
        if len(rows[0]) > len(headers):
            # Index of "Value Date" in the list
            value_date_index = headers.index('Value Date')
            headers.pop(value_date_index - 1)
            # # Insert the new item before "Value Date"
            # headers.insert(value_date_index, unknown)
            return headers

        else:
            return headers

    def get_transactions_table_rows(self, reader, page=0):
        if self.version_one:
            return self.get_transactions_table_rows_version_one(reader, page)
        if self.version_two:
            return self.get_transactions_table_rows_version_two(reader, page)
        raise UnprocessableBankStatementSupplied()

    def get_transactions_table_rows_version_one(self, reader, page=0):
        if page == 0:
            table = reader.pages[page].extract_tables()[1]
            rows_without_header = table[2:]
        else:
            table = reader.pages[page].extract_tables()[0]
            rows_without_header = table[1:]
        modified_rows = [[item.replace('\n', '').strip() if item else '' for item in row] for row in
                         rows_without_header]
        new_trans_rows = []
        pattern = r"\d{1,2}-[A-Z]{3}-\d{4}"
        second_pattern = r"\d{1,2}-[A-Z]{3}-\d{2}"
        for row in modified_rows:
            if page == 0:
                trans_date = row[0]
                value_date = row[4]
                row.pop(3)
            else:
                trans_date = row[0]
                value_date = row[3]
            if re.match(pattern, trans_date) is None:
                continue
            if re.match(second_pattern, value_date) is None:
                continue
            new_trans_rows.append(row)

        return new_trans_rows

    def read_version_one(self, reader, cleaned_text):
        account_name_extracted = self.get_account_name(cleaned_text)
        statement_period_extracted = self.get_statement_period(cleaned_text)
        account_number_extracted = self.get_account_number(cleaned_text)
        total_withdrawals_extracted = self.get_total_withdrawal(cleaned_text)
        total_deposit_extracted = self.get_total_deposit(cleaned_text)
        opening_balance_extracted = self.get_opening_balance(cleaned_text)
        closing_balance_extracted = self.get_closing_balance(cleaned_text)

        table_headers = self.get_transactions_table_headers(reader)

        num_pages = len(reader.pages)
        trans_rows = []
        for page_num in range(num_pages):
            try:
                self.set_page_state(page_num, num_pages)
                new_rows = self.get_transactions_table_rows(reader, page_num)
                if self.current_page_number == 0:
                    self.pad_header_with_unknown(new_rows, table_headers)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)
                raise e

        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df)

        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }

    def get_transactions_table_rows_version_two(self, reader, page=0):
        tables = reader.pages[page].extract_tables()
        modified_rows = []
        if self.at_first_page:
            table = tables[2]
            rows_without_header = table[1:]
        else:
            table = tables[0]
            rows_without_header = table[1:]
        for row in rows_without_header:
            if len(row) == 6:
                trans_date = row[0]
                value_date = row[1]
                transaction_details = self.clean_text(row[2])
                withdrawal = row[3]
                deposit = row[4]
                balance = row[5]
                modified_rows.append(
                    [trans_date, value_date, transaction_details, withdrawal, deposit, balance])
        return modified_rows

    def read_version_two(self, reader, cleaned_text):
        formatted_summary_table_first = self.format_account_summary_table(reader)
        formatted_summary_table_second = self.format_account_summary_table(reader, 1)
        account_name_extracted = self.get_account_name(formatted_summary_table_second)
        statement_period_extracted = self.get_statement_period(formatted_summary_table_second)
        account_number_extracted = self.get_account_number(formatted_summary_table_first)
        total_withdrawals_extracted = self.get_total_withdrawal(formatted_summary_table_first)
        total_deposit_extracted = self.get_total_deposit(formatted_summary_table_first)
        opening_balance_extracted = self.get_opening_balance(formatted_summary_table_first)
        closing_balance_extracted = self.get_closing_balance(formatted_summary_table_first)

        table_headers = self.get_transactions_table_headers(reader)
        num_pages = len(reader.pages)
        trans_rows = []

        for page_num in range(num_pages):
            try:
                self.set_page_state(page_num, num_pages)
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)
                raise e
        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df)

        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance
        }

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0 and message is not None:
            raise Exception(message)

        text = self.get_pdf_page_text(reader)
        cleaned_text = self.clean_text(text)
        first_page_tables = reader.pages[0].extract_tables()
        first_page_table_length = len(first_page_tables[0])
        if first_page_table_length == 1:
            self.version_one = True
            return self.read_version_one(reader, cleaned_text)
        if first_page_table_length == 7:
            self.version_two = True
            return self.read_version_two(reader, cleaned_text)

        raise UnprocessableBankStatementSupplied()

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]
        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            if self.version_one:
                potential_salary.append([
                    row['Transaction Date'],
                    row['Description'],
                    row['Reference'],
                    row['Value Date'],
                    row['Withdrawals'],
                    row['Deposits'],
                    row['Balance'],
                ])
            if self.version_two:
                potential_salary.append([
                    row['Transaction Date'],
                    row['Value Date'],
                    row['Description'],
                    row['Withdrawals'],
                    row['Deposits'],
                    row['Balance'],
                ])
        formatted_salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return formatted_salary_df



from bank_statement_reader.BaseBankStatementReport import BankStatementReport
import re

from bank_statement_reader.exceptions.InprocessibleBankStatementSupplied import UnprocessableBankStatementSupplied


class FirstBankStatement(BankStatementReport):
    mode = None

    def __init__(self, pdf_directory, password, min_salary, max_salary):
        if pdf_directory is None or pdf_directory == '':
            pdf_directory = "pdfs/firstbank/first_bank_version.pdf"
            # pdf_directory = "pdfs/firstbank/first_mobile_version.pdf"
            password = "34932"
        super().__init__(password=password, pdf_directory=pdf_directory, min_salary=min_salary, max_salary=max_salary,
                         bank_name='first')

    def get_opening_balance_from_table(self, reader):
        table = reader.pages[0].extract_tables()[1]
        opening_balance_row = table[1]
        last_item = opening_balance_row[-1]
        opening_balance = None
        for row in opening_balance_row:
            if 'Opening Balance' == row and self.is_money_pattern(last_item):
                opening_balance = self.convert_to_money(last_item)
                break
        return opening_balance

    def get_transactions_table_header_mapping(self):
        return {
            'transdate': 'TransDate',
            'reference': 'Reference',
            'transaction_details': 'Transaction Details',
            'valuedate': 'ValueDate',
            'deposit': 'Deposit',
            'withdrawal': 'Withdrawal',
            'balance': 'Balance'
        }

    def get_transactions_table_headers(self, reader):
        return self.get_transactions_table_header_mapping().values()

    def pad_header_with_unknown(self, rows, headers):
        if len(rows[0]) > len(headers):
            # New item to be inserted
            unknown = 'Unknown'

            # Index of "Value Date" in the list
            value_date_index = headers.index('Value Date')

            # Insert the new item before "Value Date"
            headers.insert(value_date_index, unknown)
            return headers
        else:
            return headers

    def get_transactions_table_rows_bank_version(self, reader, page=0):
        modified_rows = []
        if page == 0:
            table = reader.pages[page].extract_tables()[1]
            rows_without_header = table[1:]
            rows_without_header = rows_without_header[1:]
        else:
            table = reader.pages[page].extract_tables()[0]
            rows_without_header = table[1:]
        for row in rows_without_header:
            if len(row) == 7:
                trans_date = row[0]
                reference = row[1]
                transaction_details = row[2].replace('\n', '')
                value_date = row[3]
                withdrawal = row[4]
                deposit = row[5]
                balance = row[6]
                modified_rows.append(
                    [trans_date, reference, transaction_details, value_date, deposit, withdrawal, balance])
        if self.at_last_page is True:
            modified_rows.pop(-1)
        return modified_rows

    def get_transactions_table_rows(self, reader, page=0):
        # first_page_tables = reader.pages[0].extract_tables()
        # mobile version only has one table on first page
        if self.mode == 1:
            # print("Mobile")
            return self.get_transactions_table_rows_mobile_version(reader, page)
        # none mobile version has more than one table on first page
        if self.mode == 2:
            # print("Bank")
            return self.get_transactions_table_rows_bank_version(reader, page)

        raise UnprocessableBankStatementSupplied()

    def extract_date_by_position(self, input_string, position):
        date_formats = [
            r'\d{2}-[A-Za-z]{3}-\d{2}',
            r'\d{2}-[A-Za-z]{3}-\d{4}',
            r'\d{2}/\d{2}/\d{2}',
            r'\d{2}/\d{2}/\d{4}',
        ]
        matches = re.findall(rf'({"|".join(date_formats)})', input_string)

        if position <= len(matches):
            return matches[position - 1]
        else:
            return None

    def extract_money_by_position(self, input_string, position):
        money_formats = [
            r'\b\d{1,3}(?:,\d{3})*\.\d{2}\b',
            r'\b\d{1,3}(?:\.\d{3})*,\d{2}\b',
        ]

        matches = re.findall(rf'({"|".join(money_formats)})', input_string)
        if position <= len(matches):
            return matches[position - 1]
        else:
            return None

    def remove_substrings(self, original_string, substrings_to_remove):
        if not original_string:
            return ''
        new_string = original_string
        for substring in substrings_to_remove:
            if not substring:
                continue
            new_string = new_string.replace(substring, '')
        return new_string

    def get_transactions_table_rows_mobile_version(self, reader, page=0, on_first_page=False, on_last_page=False):
        try:
            modified_rows = []
            tables = reader.pages[page].extract_tables()
            if tables is None:
                return []
            opening_balance = None
            if page == 0:
                table = tables[0]
                rows_without_header = table[1:]
                opening_balance_row = rows_without_header[0]
                opening_balance = self.extract_money_by_position(opening_balance_row[0], 1)
                rows_without_header = rows_without_header[1:]
            else:
                rows_without_header = []
                if len(tables) > 0:
                    table = tables[0]
                    rows_without_header = table[1:]

            for row in rows_without_header:

                input_string = row[0]
                if input_string is None:
                    continue
                if input_string == '':
                    continue
                trans_date = self.extract_date_by_position(input_string, 1)
                value_date = self.extract_date_by_position(input_string, 2)
                deposit_or_withdrawal = self.extract_money_by_position(input_string, 1)
                balance = self.extract_money_by_position(input_string, 2)
                reference = ''
                transaction_details = self.remove_substrings(
                    input_string,
                    [
                        trans_date,
                        value_date,
                        deposit_or_withdrawal,
                        balance
                    ]
                )
                if transaction_details is not None:
                    transaction_details = self.clean_text(transaction_details)
                if trans_date is not None and deposit_or_withdrawal is not None and balance is not None:
                    modified_rows.append(
                        [trans_date, reference, transaction_details, value_date,
                         deposit_or_withdrawal, deposit_or_withdrawal, balance])
            for index, row in enumerate(modified_rows):
                current_row = modified_rows[index]
                current_row_balance = current_row[6]
                # row[4] is deposit column
                # row[5] is withdrawal column
                amount = row[4]
                if index == 0:
                    previous_row_balance = opening_balance
                else:
                    previous_row = modified_rows[index - 1]
                    previous_row_balance = previous_row[6]
                deposit, withdrawal = self.compute_deposit_withdrawal(current_row_balance, previous_row_balance, amount)
                if deposit is None or withdrawal is None:
                    continue
                row[4] = deposit
                row[5] = withdrawal
            return modified_rows
        except Exception as e:
            print(e)
            raise e

    def compute_deposit_withdrawal(self, current_row_balance, previous_row_balance, amount):
        if current_row_balance is None or previous_row_balance is None:
            return None, None

        if self.convert_to_money(current_row_balance) > self.convert_to_money(previous_row_balance):
            deposit = amount
            withdrawal = 0.00
        else:
            deposit = 0.00
            withdrawal = amount
        return deposit, withdrawal

    def result(self):
        reader, status, message = self.get_pdf_reader()
        print(message)
        if status == 0:
            raise Exception("Reading of file failed")
        text = self.get_pdf_page_text(reader)
        cleaned_text = self.clean_text(text)

        first_page_tables = reader.pages[0].extract_tables()

        self.mode = len(first_page_tables)
        # mobile version only has one table on first page
        if self.mode == 1:
            return self.read_mobile_app_version(reader, cleaned_text)

        # none mobile version has more than one table on first page
        if self.mode >= 2:
            return self.read_bank_version(reader, cleaned_text)

        raise UnprocessableBankStatementSupplied()

    def read_bank_version(self, reader, cleaned_text):
        num_pages = len(reader.pages)
        first_page_tables = reader.pages[0].extract_tables()
        first_page_trans_table = first_page_tables[1]
        last_page_trans_table = reader.pages[num_pages - 1].extract_tables()[0]
        opening_balance_row = first_page_trans_table[1]
        closing_balance_row = last_page_trans_table[-1]
        statement_period_extracted = self.get_statement_period(cleaned_text)
        account_name_extracted = self.get_account_name(cleaned_text)
        account_number_extracted = self.get_account_number(cleaned_text)
        total_withdrawals_extracted = self.get_total_withdrawal(cleaned_text)
        total_deposit_extracted = self.get_total_deposit(cleaned_text)
        opening_balance_extracted = self.get_opening_balance(opening_balance_row)
        closing_balance_extracted = self.get_closing_balance(closing_balance_row)
        table_headers = self.get_transactions_table_headers(reader)
        trans_rows = []
        for page_num in range(num_pages):
            try:
                self.set_page_state(page_num, num_pages)
                new_rows = self.get_transactions_table_rows(reader, page_num)
                trans_rows.extend(new_rows)
            except Exception as e:
                print(page_num)
                print("from result", e)
                raise e

        formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
        average_monthly_balance = self.get_average_monthly_balance(formatted_df)
        return {
            'dataframe': formatted_df,
            'period': statement_period_extracted,
            "account_name": account_name_extracted,
            "account_number": account_number_extracted,
            "total_turn_over_credit": total_deposit_extracted,
            "total_turn_over_debits": total_withdrawals_extracted,
            "opening_balance": opening_balance_extracted,
            "closing_balance": closing_balance_extracted,
            "average_monthly_balance": average_monthly_balance,
        }

    def read_mobile_app_version(self, reader, cleaned_text):
        try:
            statement_period_extracted = self.get_statement_period(cleaned_text)
            account_name_extracted = self.get_account_name(cleaned_text)
            account_number_extracted = self.get_account_number(cleaned_text)
            total_withdrawals_extracted = self.get_total_withdrawal(cleaned_text)
            total_deposit_extracted = self.get_total_deposit(cleaned_text)
            opening_balance_extracted = self.get_opening_balance(cleaned_text)
            closing_balance_extracted = self.get_closing_balance(cleaned_text)

            table_headers = self.get_transactions_table_headers(reader)
            num_pages = len(reader.pages)
            trans_rows = []
            for page_num in range(num_pages):

                try:
                    new_rows = self.get_transactions_table_rows(reader, page_num)
                    if page_num > 0 and len(new_rows) > 0:
                        last_row_of_trans_rows = trans_rows[-1]
                        last_row_of_new = new_rows[-1]
                        last_row_of_trans_rows_balance = last_row_of_trans_rows[-1]
                        last_row_of_new_balance = last_row_of_new[-1]
                        amount = last_row_of_new[4]
                        deposit, withdrawal = self.compute_deposit_withdrawal(last_row_of_new_balance,
                                                                              last_row_of_trans_rows_balance, amount)
                        last_row_of_new[4] = deposit
                        last_row_of_new[5] = withdrawal
                    if new_rows:
                        if len(new_rows) < 1:
                            continue
                        trans_rows.extend(new_rows)
                except Exception as e:
                    print("from result", e)
                    raise e
            formatted_df = self.format_dataframe_columns(table_headers, table_rows=trans_rows)
            # print(formatted_df)
            average_monthly_balance = self.get_average_monthly_balance(formatted_df)
            return {
                'dataframe': formatted_df,
                'period': statement_period_extracted,
                "account_name": account_name_extracted,
                "account_number": account_number_extracted,
                "total_turn_over_credit": total_deposit_extracted,
                "total_turn_over_debits": total_withdrawals_extracted,
                "opening_balance": opening_balance_extracted,
                "closing_balance": closing_balance_extracted,
                "average_monthly_balance": average_monthly_balance,
            }
        except Exception as e:
            raise e

    def predict_salary_income(self, dataframe, table_headers):
        # Filter the DataFrame to get rows with values within the specified range
        filtered_df = dataframe[(dataframe['Deposits'] >= self.min_salary) & (dataframe['Deposits'] <= self.max_salary)]
        if filtered_df.empty:
            return None

        potential_salary = []
        for index, row in filtered_df.iterrows():
            unique = self.is_unique_amount_in_month_year(row, filtered_df)
            if not unique:
                continue
            potential_salary.append([
                row['Transaction Date'],
                row['Reference'],
                row['Description'],
                row['Value Date'],
                row['Deposits'],
                row['Withdrawals'],
                row['Balance'],
            ])
        salary_df = self.format_dataframe_columns(table_headers, potential_salary)
        return salary_df
